﻿using System;
using UnityEngine;
using System.Text;
using System.Linq;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using MVR.FileManagementSecure;
using MacGruber;
using Leap.Unity;

/*

	Structure : RB-testes that has 3 colliders, c1 = rball, c2 = lball, c3 = base
	I want : 2 RBs that stay fixed with testes, with no colliders or physics, and 2 rbs with colliders and physics that are
		connected to testes with spring joints. The difference in position between an active rb and its corresponding fixed
		one will drive the morphs.
		Ideally, the active rbs should operate on the parent testes rb as well, disabling the original colliders.

		I could use the original colliders instead of the fixed RBs for getting an origin point. They will be disabled anyway.

*/



namespace UYExpBalls2
{

	

    public class UYExpBalls2 : MVRScript {
    private const string V = "uyrb";
    Transform controlL;
		Transform controlR;
		Rigidbody rigidL;
		Rigidbody rigidR;

		DAZMorph morphLFront;
		DAZMorph morphLBack;
		DAZMorph morphLLeft;
		DAZMorph morphLRight;
		DAZMorph morphLUp;
		DAZMorph morphLDown;
		DAZMorph morphRFront;
		DAZMorph morphRBack;
		DAZMorph morphRLeft;
		DAZMorph morphRRight;
		DAZMorph morphRUp;
		DAZMorph morphRDown;

		DAZMorph morphLRSqueeze;
		DAZMorph morphLRUp;
		DAZMorph morphLRDown;
		DAZMorph morphLRDownSqueeze;

		GameObject previewl;
		GameObject previewr;
		bool showpreview;
		bool seethrough = true;

		Rigidbody testesrb;
		Vector3 center_l;
		Vector3 center_r;


		Vector3 delta_l;
		Vector3 delta_r;

		JSONStorableBool showpreviewJson;

		JSONStorableFloat morphMax;
		JSONStorableFloat frontMax;
		JSONStorableFloat backMax;
		JSONStorableFloat outMax;
		JSONStorableFloat inMax;
		JSONStorableFloat upMax;
		JSONStorableFloat downMax;

		bool run;


		public override void Init() {
			
			// Atom testicleL = this.GetAtomById("TesticleLcube");
			// Atom testicleR = this.GetAtomById("TesticleRcube");

			// GameObject LGO = testicleL.GetComponentInParent<GameObject>();


			var rigidbodies = containingAtom.GetComponentsInChildren<Rigidbody>();
			testesrb = null;
			foreach (var r in rigidbodies) {
				if (r.name=="Testes") testesrb = r;
			}

			SuperController.LogMessage($" found testesrb {testesrb!=null} {testesrb.mass}");

			CreateOrSyncTesticles();

			// controlL = testicleL.mainController.transform;
			// controlR = testicleR.mainController.transform;
			// rigidL = testicleL.realRigidbodies[0].transform;
			// rigidR = testicleR.realRigidbodies[0].transform;

			var morphs = returnMorphs();
			morphLFront = morphs.GetMorphByUid("UrukYay.UYBalls.1:/Custom/Atom/Person/Morphs/male_genitalia/UYBalls/testicleLFront.vmi");
			morphLBack = morphs.GetMorphByUid("UrukYay.UYBalls.1:/Custom/Atom/Person/Morphs/male_genitalia/UYBalls/testicleLBack.vmi");
			morphLLeft = morphs.GetMorphByUid("UrukYay.UYBalls.1:/Custom/Atom/Person/Morphs/male_genitalia/UYBalls/testicleLLeft.vmi");
			morphLRight = morphs.GetMorphByUid("UrukYay.UYBalls.1:/Custom/Atom/Person/Morphs/male_genitalia/UYBalls/testicleLRight.vmi");
			morphLUp = morphs.GetMorphByUid("UrukYay.UYBalls.1:/Custom/Atom/Person/Morphs/male_genitalia/UYBalls/testicleLUp.vmi");
			morphLDown = morphs.GetMorphByUid("UrukYay.UYBalls.1:/Custom/Atom/Person/Morphs/male_genitalia/UYBalls/testicleLDown.vmi");

			morphRFront = morphs.GetMorphByUid("UrukYay.UYBalls.1:/Custom/Atom/Person/Morphs/male_genitalia/UYBalls/testicleRFront.mtl.vmi");
			morphRBack = morphs.GetMorphByUid("UrukYay.UYBalls.1:/Custom/Atom/Person/Morphs/male_genitalia/UYBalls/testicleRBack.vmi");
			morphRLeft = morphs.GetMorphByUid("UrukYay.UYBalls.1:/Custom/Atom/Person/Morphs/male_genitalia/UYBalls/testicleRLeft.vmi");
			morphRRight = morphs.GetMorphByUid("UrukYay.UYBalls.1:/Custom/Atom/Person/Morphs/male_genitalia/UYBalls/testicleRRight.vmi");
			morphRUp = morphs.GetMorphByUid("UrukYay.UYBalls.1:/Custom/Atom/Person/Morphs/male_genitalia/UYBalls/testicleRUp.vmi");
			morphRDown = morphs.GetMorphByUid("UrukYay.UYBalls.1:/Custom/Atom/Person/Morphs/male_genitalia/UYBalls/testicleRDown.vmi");

			morphLRSqueeze = morphs.GetMorphByUid("UrukYay.UYBalls.1:/Custom/Atom/Person/Morphs/male_genitalia/UYBalls/testicleLRSqueezeCorrective.vmi");
			morphLRUp = morphs.GetMorphByUid("UrukYay.UYBalls.1:/Custom/Atom/Person/Morphs/male_genitalia/UYBalls/testicleLRUpCorrective.vmi");
			morphLRDown = morphs.GetMorphByUid("UrukYay.UYBalls.1:/Custom/Atom/Person/Morphs/male_genitalia/UYBalls/testicleLRDownCorrective.vmi");
			morphLRDownSqueeze = morphs.GetMorphByUid("UrukYay.UYBalls.1:/Custom/Atom/Person/Morphs/male_genitalia/UYBalls/testicleLRDownSqueezeCorrective.vmi");

			// EnablePreviews(true);
			ShowUI();
			run = true;
		}


		public void CreateOrSyncTesticles() {
			float right_radius = 0;
			float left_radius = 0;
			float right_height = 0;
			float left_height = 0;
			Vector3 right_center = new Vector3();
			Vector3 left_center = new Vector3();

			var colliders = containingAtom.GetComponentsInChildren<CapsuleCollider>();
			foreach (var c in colliders) {
				if (c.attachedRigidbody == testesrb ) {
					SuperController.LogMessage($"found col {c.name} {c.height} {c.radius} {c.center.ToString("F3")}");
					SuperController.LogMessage($"{c.transform.localPosition.ToString("F3")}  {testesrb.transform.localPosition.ToString("F3")}");
					if (c.name == "_Collider1") {
						right_height=c.height;
						right_radius=c.radius;
						c.isTrigger=true;
						right_center = c.transform.localPosition + c.center;
						controlR = c.transform;
						center_r = c.center;
					} else if (c.name == "_Collider2") {
						left_height=c.height;
						left_radius=c.radius;
						c.isTrigger=true;
						left_center = c.transform.localPosition + c.center;
						controlL = c.transform;
						center_l = c.center;
					} else if (c.name == "_Collider3") {
						c.isTrigger = false;
						foreach (var c2 in colliders) {
							if (c2.attachedRigidbody.name == "jointB1Male") {
								Physics.IgnoreCollision(c, c2);
							} else if (c2.attachedRigidbody.name == "jointB2Male") {
								Physics.IgnoreCollision(c, c2);
							} else if (c2.attachedRigidbody.name == "jointAlMale") {
								Physics.IgnoreCollision(c, c2);
							} else if (c2.attachedRigidbody.name == "jointArMale") {
								Physics.IgnoreCollision(c, c2);
							} else if (c2.attachedRigidbody.name.Equals("lThigh")) {
								SuperController.LogMessage($"ignoring lthigh collision for {c2.name}");
								Physics.IgnoreCollision(c, c2);
							} else if (c2.attachedRigidbody.name == "rThigh") {
								SuperController.LogMessage($"ignoring rthigh collision for {c2.name}");
								Physics.IgnoreCollision(c, c2);
							}							

						}
					}
					
				}
			}

			testesrb.centerOfMass = (left_center + right_center) / 2f;

			rigidL = CreateTesticle("uy_newtesticleL", left_height, left_radius, left_center);

			rigidR = CreateTesticle("uy_newtesticleR", right_height, right_radius, right_center);

			CapsuleCollider tlc = rigidL.GetComponent<CapsuleCollider>();
			CapsuleCollider trc = rigidR.GetComponent<CapsuleCollider>();

			Physics.IgnoreCollision(tlc,trc);


		}

		public Rigidbody CreateTesticle ( string name, float height, float radius, Vector3 center) {
			Rigidbody newrb = null;
			GameObject newgo = GameObject.Find(name);
			CapsuleCollider trc = null;
			UnityEngine.ConfigurableJoint j = null;
			if (newgo==null) {
				SuperController.LogMessage($"Creating new go rb and cc");
				// DestroyImmediate(newgo);
				newgo = new GameObject();
				newgo.name = name;
				newrb = newgo.AddComponent<Rigidbody>();
				trc = newgo.AddComponent<CapsuleCollider>();
				j = newgo.AddComponent<UnityEngine.ConfigurableJoint>();
			}

			// SuperController.LogMessage($"Creating new go rb and cc");
			
			newgo.transform.SetParent(testesrb.transform, false);
			newrb = newgo.GetComponent<Rigidbody>();
			newrb.transform.localPosition = center;
			trc = newgo.GetComponent<CapsuleCollider>();
			j = newgo.GetComponent<UnityEngine.ConfigurableJoint>();
			SuperController.LogMessage ($"found newrb {newrb!=null} trc {trc!=null} j {j!=null}");
			trc.height = height;
			trc.radius = radius;
			trc.direction = 2;	
			
			newrb.isKinematic = false;
			newrb.mass = 0.01f;
			newrb.useGravity = true;
			// j.transform.localPosition = Vector3.zero;
			j.anchor = Vector3.zero;
			j.autoConfigureConnectedAnchor=true;
			j.connectedBody = testesrb;

			j.xMotion = ConfigurableJointMotion.Limited;
			j.yMotion = ConfigurableJointMotion.Limited;
			j.zMotion = ConfigurableJointMotion.Limited;
			j.angularXMotion = ConfigurableJointMotion.Locked;
			j.angularYMotion = ConfigurableJointMotion.Locked;
			j.angularZMotion = ConfigurableJointMotion.Locked;

			JointDrive jd = j.xDrive;
			jd.positionSpring = 15;
			jd.positionDamper = 1f;
			// jd.maximumForce = 10000000f;
			j.xDrive = jd;
			j.yDrive = jd;
			j.zDrive = jd;
			j.targetRotation = Quaternion.identity;
			j.targetPosition = Vector3.zero;
			
			// JointDrive jsd = j.slerpDrive;
			// jsd.positionSpring = 0;
			// jsd.positionDamper = 0.1f;
			// jsd.maximumForce = 100000000;
			// j.slerpDrive = jsd;

			SoftJointLimit jl = j.linearLimit;
			jl.limit=0.05f;
			jl.bounciness=0f;
			j.linearLimit = jl;
			SoftJointLimitSpring jls = j.linearLimitSpring;
			jls.spring = 0f;
			jls.damper = 0f;
			j.linearLimitSpring = jls;
			// j.projectionMode = JointProjectionMode.PositionAndRotation;
			j.projectionMode = JointProjectionMode.None;

			SuperController.LogMessage($"testesrb : {testesrb.position}  uy : {newrb.position} {trc.center.ToString("F3")} {trc.transform.localPosition.ToString("F3")}");

			return newrb;
		}

		public void ResetToOriginal () {
			SuperController.LogMessage($"resetting {containingAtom==null}");

			var colliders = containingAtom.GetComponentsInChildren<CapsuleCollider>();
			SuperController.LogMessage($"got colliders");
			foreach (var c in colliders) {
				if (c.attachedRigidbody == testesrb ) {
					SuperController.LogMessage($"resetting collider {c.name}");

					if (c.name == "_Collider1") {
						c.isTrigger=false;
					} else if (c.name == "_Collider2") {
						c.isTrigger=false;
					} else if (c.name == "_Collider3") {
						c.isTrigger = false;
						foreach (var c2 in colliders) {
							if (c2.attachedRigidbody.name == "jointB1Male") {
								Physics.IgnoreCollision(c, c2, false);
							} else if (c2.attachedRigidbody.name == "jointB2Male") {
								Physics.IgnoreCollision(c, c2, false);
							} else if (c2.attachedRigidbody.name == "jointAlMale") {
								Physics.IgnoreCollision(c, c2, false);
							} else if (c2.attachedRigidbody.name == "jointArMale") {
								Physics.IgnoreCollision(c, c2, false);
							} else if (c2.attachedRigidbody.name.Equals("lThigh")) {
								Physics.IgnoreCollision(c, c2, false);
							} else if (c2.attachedRigidbody.name == "rThigh") {
								Physics.IgnoreCollision(c, c2, false);
							}							

						}
					}
					
				}
			}			
			SuperController.LogMessage($"reset colliders");

			testesrb.ResetCenterOfMass();
			Destroy(rigidL.gameObject);
			Destroy(rigidR.gameObject);
			// rigidL = null;
			// rigidR = null;
			SuperController.LogMessage($"destroyed objects");
			morphLFront.morphValue = 0f;
			morphLBack.morphValue = 0f;
			morphLUp.morphValue = 0f;
			morphLDown.morphValue = 0f;
			morphLLeft.morphValue = 0f;
			morphLRight.morphValue = 0f;			
			morphRFront.morphValue = 0f;
			morphRBack.morphValue = 0f;
			morphRUp.morphValue = 0f;
			morphRDown.morphValue = 0f;
			morphRLeft.morphValue = 0f;
			morphRRight.morphValue = 0f;
			morphLRDown.morphValue = 0f;
			morphLRUp.morphValue = 0f;
			morphLRSqueeze.morphValue = 0f;
			morphLRDownSqueeze.morphValue = 0f;
			SuperController.LogMessage($"reset morphs");
			run = false;
		}

		public void ShowUI()
		{
			showpreviewJson = new JSONStorableBool("show preview", false, EnablePreviews);
			CreateToggle(showpreviewJson);
			MacGruber.Utils.SetupButton(this, "Sync To Colliders", CreateOrSyncTesticles, false);

			morphMax= new JSONStorableFloat("Morph max value", 2.0f, 0f, 1f, false);
			frontMax= new JSONStorableFloat("Front max distance", 0.05f, 0f, 0.5f, false);			
			backMax = new JSONStorableFloat("Back max distance", 0.05f, 0f, 0.5f, false);			
			upMax 	= new JSONStorableFloat("Up max distance", 0.05f, 0f, 0.25f, false);			
			downMax = new JSONStorableFloat("Down max distance", 0.08f, 0f, 0.5f, false);			
			outMax 	= new JSONStorableFloat("Out (side/side) max distance", 0.07f, 0f, 0.5f, false);			
			inMax 	= new JSONStorableFloat("In (side/side) max distance", 0.04f, 0f, 0.25f, false);			

			RegisterFloat(morphMax);
			RegisterFloat(frontMax);
			RegisterFloat(backMax);
			RegisterFloat(upMax);
			RegisterFloat(downMax);
			RegisterFloat(outMax);
			RegisterFloat(inMax);

			CreateSlider(morphMax);
			CreateSlider(frontMax);
			CreateSlider(backMax);
			CreateSlider(upMax);
			CreateSlider(downMax);
			CreateSlider(outMax);
			CreateSlider(inMax);

			// adjust_x_l = new JSONStorableFloat("Adjust X left", 0f, UpdateAdjustVector, -0.1f, 0.1f,false,true);
			// adjust_y_l = new JSONStorableFloat("Adjust Y left", 0f, UpdateAdjustVector, -0.1f, 0.1f,false);
			// adjust_z_l = new JSONStorableFloat("Adjust Z left", 0f, UpdateAdjustVector, -0.1f, 0.1f,false);

			// // adjust_rad_l = new JSONStorableFloat("Adjust Radius left", tlc.radius, UpdateAdjustVector, tlc.radius*0.7f, tlc.radius*1.3f, false);
			// // adjust_h_l = new JSONStorableFloat("Adjust Height left", tlc.height, UpdateAdjustVector, tlc.height*0.7f, tlc.height*1.3f, false);


			// adjust_x_r = new JSONStorableFloat("Adjust X right", 0f, UpdateAdjustVector, -0.1f, 0.1f,false);
			// adjust_y_r = new JSONStorableFloat("Adjust Y right", 0f, UpdateAdjustVector, -0.1f, 0.1f,false);
			// adjust_z_r = new JSONStorableFloat("Adjust Z right", 0f, UpdateAdjustVector, -0.1f, 0.1f,false);

			// adjust_rad_r = new JSONStorableFloat("Adjust Radius right", trc.radius, UpdateAdjustVector, trc.radius*0.7f, trc.radius*1.3f, false);
			// adjust_h_r = new JSONStorableFloat("Adjust Height right", trc.height, UpdateAdjustVector, trc.height*0.7f, trc.height*1.3f, false);


			// RegisterFloat(adjust_x_l);
			// RegisterFloat(adjust_y_l);
			// RegisterFloat(adjust_z_l);
			// // RegisterFloat(adjust_rad_l);
			// // RegisterFloat(adjust_h_l);
			// RegisterFloat(adjust_x_r);
			// RegisterFloat(adjust_y_r);
			// RegisterFloat(adjust_z_r);
			// RegisterFloat(adjust_rad_r);
			// RegisterFloat(adjust_h_r);

			// CreateSlider(adjust_x_l, true);
			// CreateSlider(adjust_y_l, true);
			// CreateSlider(adjust_z_l, true);
			// // CreateSlider(adjust_rad_l, true);
			// // CreateSlider(adjust_h_l, true);
			// CreateSpacer(true);
			// CreateSlider(adjust_x_r, true);
			// CreateSlider(adjust_y_r, true);
			// CreateSlider(adjust_z_r, true);
			// CreateSlider(adjust_rad_r, true);
			// CreateSlider(adjust_h_r, true);


		}

		// void UpdateAdjustVector(float f)
		// {
		// 	adjust_l.x= -0.0135f - 0.005f - adjust_x_l.val;
		// 	adjust_l.y = -0.1041f - 0.015f + adjust_y_l.val;
		// 	adjust_l.z = adjust_z_l.val;
		// 	adjust_r.x= 0.0135f + 0.005f + adjust_x_r.val;
		// 	adjust_r.y = -0.1041f - 0.015f + adjust_y_r.val;
		// 	adjust_r.z = adjust_z_r.val;

		// 	// tlc.radius = adjust_rad_l.val;
		// 	// tlc.height = adjust_h_l.val;
		// 	// trc.radius = adjust_rad_r.val;
		// 	// trc.height = adjust_h_r.val;
		// 	// UpdatePreview(previewl, tlc);
		// 	UpdatePreview(previewr, trc);

		// }

		void EnablePreviews( bool val) {
			if (val) {
				showpreview = true;
				CapsuleCollider tlc = rigidL.GetComponent<CapsuleCollider>();
				CapsuleCollider trc = rigidR.GetComponent<CapsuleCollider>();
				previewl = createPreview(tlc);
				previewr = createPreview(trc);

			} else {
				if (showpreview) {
					destroyPreview(previewl);
					destroyPreview(previewr);
					showpreview = false;
				}
			}
		}

		void Start() {
		}
		void Update() {

			// if (!run) return;


			// delta_l = controlL.position + center_l - rigidL.transform.position;

			// delta_l = Quaternion.Inverse(rigidL.transform.rotation) * delta_l;

			delta_l = controlL.localPosition + center_l - rigidL.transform.localPosition;

			delta_l = Quaternion.Inverse(rigidL.transform.localRotation) * delta_l;



			float upvall = (delta_l.y<=0) ? morphMax.val * Mathf.Clamp01(-delta_l.y/upMax.val) : 0f;
			float downvall = (delta_l.y>0) ? morphMax.val * Mathf.Clamp01(delta_l.y/downMax.val) : 0f;

			float rightvall = (delta_l.x<=0) ? morphMax.val * Mathf.Clamp01(-delta_l.x/inMax.val) : 0f;
			float leftvall = (delta_l.x>0) ? morphMax.val * Mathf.Clamp01(delta_l.x/outMax.val) : 0f;

			float frontvall = (delta_l.z<=0) ? morphMax.val * Mathf.Clamp01(-delta_l.z/frontMax.val) : 0f;
			float backvall = (delta_l.z>0) ? morphMax.val * Mathf.Clamp01(delta_l.z/backMax.val) : 0f;

			morphLFront.morphValue = frontvall;
			morphLBack.morphValue = backvall;
			morphLUp.morphValue = upvall;
			morphLDown.morphValue = downvall;
			morphLLeft.morphValue = leftvall;
			morphLRight.morphValue = rightvall;

			
			// delta_r = controlR.position + center_r - rigidR.transform.position;

			// delta_r = Quaternion.Inverse(rigidR.transform.rotation) * delta_r;

			delta_r = controlR.localPosition + center_r - rigidR.transform.localPosition;

			delta_r = Quaternion.Inverse(rigidR.transform.localRotation) * delta_r;


			float upvalr = (delta_r.y<=0) ? morphMax.val * Mathf.Clamp01(-delta_r.y/upMax.val) : 0f;
			float downvalr = (delta_r.y>0) ? morphMax.val * Mathf.Clamp01(delta_r.y/downMax.val) : 0f;

			float rightvalr = (delta_r.x<=0) ? morphMax.val * Mathf.Clamp01(-delta_r.x/outMax.val) : 0f;
			float leftvalr = (delta_r.x>0) ? morphMax.val * Mathf.Clamp01(delta_r.x/inMax.val) : 0f;

			float frontvalr = (delta_r.z<=0) ? morphMax.val * Mathf.Clamp01(-delta_r.z/frontMax.val) : 0f;
			float backvalr = (delta_r.z>0) ? morphMax.val * Mathf.Clamp01(delta_r.z/backMax.val) : 0f;

			morphRFront.morphValue = frontvalr;
			morphRBack.morphValue = backvalr;
			morphRUp.morphValue = upvalr;
			morphRDown.morphValue = downvalr;
			morphRLeft.morphValue = leftvalr;
			morphRRight.morphValue = rightvalr;


			morphLRSqueeze.morphValue = (rightvall + leftvalr)/2.0f;
			morphLRUp.morphValue = (upvall + upvalr)/2.0f;

			// down corrective : 1.6 * avglr - 0.6
			morphLRDown.morphValue = Mathf.Clamp01(0.8f*(downvall + downvalr) - 0.6f);
			morphLRDownSqueeze.morphValue = (morphLRSqueeze.morphValue + morphLRDown.morphValue) / 2.0f;

			// SuperController.LogMessage($"R {delta_r.ToString("F4")}, {delta_l.ToString("F4")} ");



		}
		void FixedUpdate() {
			
		}
    public void OnEnable()
    {
				CreateOrSyncTesticles();

        try
        {
            SuperController.LogMessage($"{nameof(UYExpBalls2)} enabled");
        }
        catch (Exception e)
        {
            SuperController.LogError($"{nameof(UYExpBalls2)}.{nameof(OnEnable)}: {e}");
        }
    }

    public void OnDisable()
    {

				destroyPreview(previewl);
				destroyPreview(previewr);
				ResetToOriginal();
        try
        {
            SuperController.LogMessage($"{nameof(UYExpBalls2)} disabled");
        }
        catch (Exception e)
        {
            SuperController.LogError($"{nameof(UYExpBalls2)}.{nameof(OnDisable)}: {e}");
        }
    }

    public void OnDestroy()
   
    {

        try
        {
            SuperController.LogMessage($"{nameof(UYExpBalls2)} destroyed");
        }
        catch (Exception e)
        {
            SuperController.LogError($"{nameof(UYExpBalls2)}.{nameof(OnDestroy)}: {e}");
        }
    }


		GenerateDAZMorphsControlUI returnMorphs() {
			JSONStorable geometry = containingAtom.GetStorableByID("geometry");
			DAZCharacterSelector character = geometry as DAZCharacterSelector;
			return character.morphsControlUI;
		}


	private Mesh makeTubeMesh()
	{
			Mesh m = new Mesh();
			Vector3 [] vertices = new Vector3[24];   
			for (int i=0; i<12; i++)
			{
					float r=0.5f;
					vertices[i] = new Vector3( r*Mathf.Cos(i*2*Mathf.PI/12f), -1, r*Mathf.Sin(i*2*Mathf.PI/12f));
			}
			for (int i=0; i<12; i++)
			{
					float r=0.5f;
					vertices[i+12] = new Vector3( r*Mathf.Cos(i*2*Mathf.PI/12f), 1, r*Mathf.Sin(i*2*Mathf.PI/12f));
			}

			int[] triangles = new int[72];
			for (int i=0;i<12; i++)
			{
					triangles[i*3] = i;
					triangles[i*3+1] = i+12;
					triangles[i*3+2] = (i+1) % 12;
			}
			for (int i=0;i<12; i++)
			{
					triangles[i*3+36] = (i+1) % 12;
					triangles[i*3+37] = i+12;
					triangles[i*3+38] = (i+1) % 12 + 12;
			}

			m.vertices = vertices;
			m.triangles = triangles;
			m.RecalculateNormals();
			m.RecalculateBounds();
			
			return m;
	}

	private Mesh makeHemisphereMesh()
	{
			Mesh m = new Mesh();
			Vector3 [] vertices = new Vector3[37];   
			for (int i=0; i<12; i++)
			{
					float h = 0;
					float r = 1f;
					vertices[i] = new Vector3( r*Mathf.Cos(i*2*Mathf.PI/12f)/2, h/2, r*Mathf.Sin(i*2*Mathf.PI/12f)/2);
			}
			for (int i=0; i<12; i++)
			{
					float h = Mathf.Sin(Mathf.PI/6);
					float r = Mathf.Cos(Mathf.PI/6);
					vertices[i+12] = new Vector3( r*Mathf.Cos(i*2*Mathf.PI/12f)/2, h/2, r*Mathf.Sin(i*2*Mathf.PI/12f)/2);
			}
			for (int i=0; i<12; i++)
			{
					float h = Mathf.Sin(Mathf.PI/3);
					float r = Mathf.Cos(Mathf.PI/3);
					vertices[i+24] = new Vector3( r*Mathf.Cos(i*2*Mathf.PI/12f)/2, h/2, r*Mathf.Sin(i*2*Mathf.PI/12f)/2);
			}
			vertices[36] = new Vector3(0,0.5f,0);


			int[] triangles = new int[180];
			for (int i=0;i<12; i++) // bottom row  /\ 0-35
			{
					triangles[i*3] = i;
					triangles[i*3+1] = i+12;
					triangles[i*3+2] = (i+1) % 12;
			}
			for (int i=0;i<12; i++) // bottom row \/ 36-71
			{
					triangles[i*3+36] = (i+1) % 12;
					triangles[i*3+37] = i+12;
					triangles[i*3+38] = (i+1) % 12 + 12;
			}

			for (int i=0;i<12; i++) // mid row /\ 72-107
			{
					triangles[i*3+72] = i+12;
					triangles[i*3+73] = i+24;
					triangles[i*3+74] = (i+1) % 12 + 12;
			}
			for (int i=0;i<12; i++) // mid row \/ 108-143
			{
					triangles[i*3+108] = (i+1) % 12 + 12;
					triangles[i*3+109] = i+24;
					triangles[i*3+110] = (i+1) % 12 + 24;
			}

			for (int i=0;i<12; i++) // top row /\ 144-179
			{
					triangles[i*3+144] = i+24;
					triangles[i*3+145] = 36;
					triangles[i*3+146] = (i+1) % 12 + 24;
			}



			m.vertices = vertices;
			m.triangles = triangles;
			m.RecalculateNormals();
			m.RecalculateBounds();
			
			return m;
	}


		public GameObject createPreview(CapsuleCollider ccollider)
		{
				// preview=GameObject.CreatePrimitive(PrimitiveType.Cylinder);
				// GameObject sphere1 = GameObject.CreatePrimitive(PrimitiveType.Sphere);
				// sphere1.transform.position = preview.transform.position + Vector3.up;
				// sphere1.transform.SetParent(preview.transform,true);
				GameObject preview = new GameObject();
				MeshFilter mf = preview.AddComponent<MeshFilter>() as MeshFilter;
				preview.AddComponent<MeshRenderer>();
				mf.sharedMesh = makeTubeMesh();

				GameObject sphere1 = new GameObject();
				mf = sphere1.AddComponent<MeshFilter>() as MeshFilter;
				sphere1.AddComponent<MeshRenderer>();
				mf.sharedMesh = makeHemisphereMesh();
				sphere1.transform.position = preview.transform.position + Vector3.up;
				sphere1.transform.SetParent(preview.transform,true);


				GameObject sphere2 = new GameObject();
				mf = sphere2.AddComponent<MeshFilter>() as MeshFilter;
				sphere2.AddComponent<MeshRenderer>();
				mf.sharedMesh = makeHemisphereMesh();
				sphere2.transform.position = preview.transform.position - Vector3.up;
				sphere2.transform.Rotate(new Vector3(180,0,0));
				sphere2.transform.SetParent(preview.transform,true);



				preview.transform.SetParent(ccollider.transform);
				foreach (var c in preview.GetComponentsInChildren<Collider>())
				{
						c.enabled = false;
						UnityEngine.Object.Destroy(c);
				}

				// setInactive();
				UpdatePreview(preview, ccollider);
				return preview;

		}
		public void destroyPreview(GameObject preview)
		{
				UnityEngine.Object.Destroy(preview);
		}

		public void UpdatePreview(GameObject preview, CapsuleCollider ccollider)
		{
				if (showpreview) {
					var allPreviewRenderers = preview.GetComponentsInChildren<Renderer>(); 
						// SuperController.LogMessage($"renderers : {previewrenderers.Length} ");
						// var previewRenderer = preview.GetComponent<Renderer>();
					foreach (var previewRenderer in allPreviewRenderers)
					{
						var material = previewRenderer.material;
						if (seethrough)
						{
								material.shader = Shader.Find("Battlehub/RTGizmos/Handles");
								material.SetFloat("_Offset", 1f);
								material.SetFloat("_MinAlpha", 1f);
								previewRenderer.material = material;
						}
						else
						{
								material.shader = Shader.Find("Standard");
								material.SetInt("_SrcBlend", (int)UnityEngine.Rendering.BlendMode.SrcAlpha);
								material.SetInt("_DstBlend", (int)UnityEngine.Rendering.BlendMode.OneMinusSrcAlpha);
								material.SetInt("_ZWrite", 0);
								material.DisableKeyword("_ALPHATEST_ON");
								material.EnableKeyword("_ALPHABLEND_ON");
								material.DisableKeyword("_ALPHAPREMULTIPLY_ON");
								material.renderQueue = 3000;
								previewRenderer.material = material;
						}

						float size = ccollider.radius * 2;
						float height = ccollider.height / 2 - size / 2;
						if (height==0f) height = 0.00001f;
						preview.transform.localScale = new Vector3(size, height, size);
						preview.transform.GetChild(0).localScale = new Vector3(1, size/height, 1);
						preview.transform.GetChild(1).localScale = new Vector3(1, size/height, 1);
						if (ccollider.direction == 0)
								preview.transform.localRotation = Quaternion.AngleAxis(90, Vector3.forward);
						else if (ccollider.direction == 2)
								preview.transform.localRotation = Quaternion.AngleAxis(90, Vector3.right);
						preview.transform.localPosition = ccollider.center;

						var color = Color.green;
						color.a = seethrough ? 0.5f : 0.5f;
						previewRenderer.material.color = color;

						// if (gui) {
						// 		// var previewRenderer = preview.GetComponent<Renderer>();
						// 		// var material = previewRenderer.material;
						// 		var color = Color.magenta;//previewRenderer.material.color;
						// 		// color.a = seethrough ? 0.53f : 0.67f;
						// 		color.a = seethrough ? Mathf.Clamp01(seethroughalpha * 10) : Mathf.Clamp01(occludedalpha*2);
						// 		previewRenderer.material.color = color;
						// } 
						// else
						// {
						// 		// var previewRenderer = preview.GetComponent<Renderer>();
						// 		// var material = previewRenderer.material;
						// 		var color = interp_n1!=null && interp_n2!=null ? Color.yellow : Color.green;//previewRenderer.material.color;
						// 		// color.a = seethrough ? 0.05f : 0.3f;
						// 		color.a = seethrough ? seethroughalpha : occludedalpha;
						// 		previewRenderer.material.color = color;                    
						// }
					}
				}
		}



	}
}